/* gmt2cha.c */
/* Convert GMT postscript maps to CHART vector charts. */

/* Author:  Heiko Purnhagen */
/* e-mail:  purnhage@tnt.uni-hannover.de */
/* WWW:     http://www.fet.uni-hannover.de/purnhage/ */
/* Copying: GNU General Public Licence */
/* Changes: 970703 970704 */

/* GMT (Generic Mapping Tools) is available at: */
/* http://www.soest.hawaii.edu/soest/gmt.html */
/* ftp://kiawe.soest.hawaii.edu/ */
/* ftp://ibis.grdl.noaa.gov/ */
/* ftp://ftp.geologi.uio.no/ */

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>


#define PROGNAME "gmt2cha"
#define PROGVER PROGNAME " 1.0 970704"

#define STRLEN 256
#define BUFSIZE 10000
#define SCALELIMIT 20
#define NAUT_MILE (40000000./21600.)

#define max(a,b) ((a) > (b) ? (a) : (b))


FILE *in,*out;
long gmtline,gmtpoly,gmtpoint;
long chaline,chapoly,chapoint;
long psline;
char *section;

double llw,lle,lls,lln,width,psunit;
double scalex,scaley;
int skiprate;

int bufidx;
int polyflag;
double buflat[BUFSIZE];
double buflon[BUFSIZE];


void errorexit (int errorcode, char *message, ...)
{
  va_list args;

  va_start(args,message);
  fflush(stdout);
  fprintf(stderr,"%s: ERROR[%d]: ",PROGNAME,errorcode);
  vfprintf(stderr,message,args);
  fprintf(stderr,"\n");
  va_end(args);
  exit (errorcode);
}


char *lastchar (char *s)
{
  if (*s) {
    while (*s)
      s++;
    s--;
  }
  return s;
}


void getl (char *s)
{
  char *p;

  if (!fgets(s,STRLEN,in))
    errorexit(1,"unexpected end of input (line %d / %s)",psline,section);
  psline++;
  if (*(p=lastchar(s))!='\n')
    errorexit(1,"line too long (line %d / %s)",psline,section);
  *p = '\0';
}


char *getdbl (char *s, char *key, double *d)
{
  if (!(s=strstr(s,key)))
    return key;
  s += strlen(key);
  if (sscanf(s,"%lf",d)!=1)
    return key;
  return (char*)NULL;
}


int getang (char *s, double *d)
{
  char *p;
  double m;

  if ((p=strchr(s,':')))
    *p++ = '\0';
  if (sscanf(s,"%lf",d)!=1)
    return 1;
  if (p) {
    if (sscanf(p,"%lf",&m)!=1)
      return 1;
    if (*d>=0)
      *d += m/60.;
    else
      *d -= m/60.;
  }
  return 0;
}


char *getwesn (char *s, double *llw, double *lle, double *lls, double *lln)
{
  char *p;

  if (!(s=strstr(s,"-R")))
    return "-R";
  s += 2;
  p = s;
  while (*p)
    if (*p==' ' || *p=='\t')
      *p = '\0';
    else
      p++;

  if (!(p=strchr(s,'/')))
    return "west";
  *p++ = '\0';
  if (getang(s,llw))
    return "west";
  s = p;
  if (!(p=strchr(s,'/')))
    return "east";
  *p++ = '\0';
  if (getang(s,lle))
    return "east";
  s = p;
  if (!(p=strchr(s,'/')))
    return "south";
  *p++ = '\0';
  if (getang(s,lls))
    return "south";
  s = p;
  if (getang(s,lln))
    return "north";
  
  return (char*)NULL;
}


void prline (char *fmt, ...)
{
  va_list args;

  va_start(args,fmt);
  vfprintf(out,fmt,args);
  fprintf(out,"\r\n");
  va_end(args);
}


char *latlonstr (char *s, double lat, double lon)
{
  int latd,lond;
  double latm,lonm;
  char latc,lonc;

  latc = 'N';
  if (lat<0) {
    lat = -lat;
    latc = 'S';
  }
  latd = (int)lat;
  latm = (lat-latd)*60;
  sprintf(s,"%9.6f",latm);
  if (*s=='6') {
    latd++;
    latm = 0;
  }
  lonc = 'E';
  if (lon<0) {
    lon = -lon;
    lonc = 'W';
  }
  lond = (int)lon;
  lonm = (lon-lond)*60;
  sprintf(s,"%9.6f",lonm);
  if (*s=='6') {
    lond++;
    lonm = 0;
  }
  sprintf(s,"%2d %9.6f %c %3d %9.6f %c",latd,latm,latc,lond,lonm,lonc);
  if (s[3]==' ')
    s[3] = '0';
  if (s[19]==' ')
    s[19] = '0';
  return s;
}


void bufadd (long x, long y)
{
  double lat,lon;

  if (bufidx==BUFSIZE)
    errorexit(5,"line/poly too long (line %d)",psline);
  lat = y*scaley+lls;
  lon = x*scalex+llw;
  if (lat<lls-scaley || lat>lln+scaley || lon<llw-scalex || lon>lle+scalex)
    errorexit(5,"lat/lon out off range (line %d)",psline);
  buflat[bufidx] = lat;
  buflon[bufidx] = lon;
  bufidx++;
  gmtpoint++;
}

void bufwrite (void)
{
  int i;
  char s[STRLEN];

  if (polyflag)
    gmtpoly++;
  else
    gmtline++;
  if (!polyflag || bufidx>=skiprate) {
    if (polyflag)
      chapoly++;
    else
      chaline++;
    prline("%s %9.3f",
	   polyflag?"poly":"line",
	   max(lle-llw,lln-lls)*60.*SCALELIMIT);
    prline("GMT");
    for (i=0; i<bufidx; i+=skiprate) {
      prline("%s",latlonstr(s,buflat[i],buflon[i]));
      chapoint++;
    }
    if (i-skiprate != bufidx-1) {
      prline("%s",latlonstr(s,buflat[bufidx-1],buflon[bufidx-1]));
      chapoint++;
    }
  }
  bufidx = 0;
  polyflag = 0;
}


void convert (void)
{
  char s[STRLEN];
  char *err;
  char *p,*n;
  char c;
  long x,y,dx,dy;
  int state;

  psline = 0;

  /* header */
  prline(". %s",PROGVER);

  section = "%!PS-Adobe";
  do {
    getl(s);
  } while (!strstr(s,section));

  section = "Postscript units";
  do {
    getl(s);
  } while (!strstr(s,section));
  prline(". %s",s+2);
  if ((err=getdbl(s+2,"on the paper equals ",&psunit)))
    errorexit(2,"postscript units parse error (%s) (line %d)",err,psline);

  section = "pscoast";
  do {
    getl(s);
  } while (!strstr(s,section));
  prline(". %s",s+2);
  if ((err=getdbl(s+2,"-JX",&width)))
    errorexit(2,"postscript units parse error (%s) (line %d)",
	      err,psline);
  if ((err=getwesn(s+2,&llw,&lle,&lls,&lln)))
    errorexit(2,"pscoast option parse error (%s) (line %d)",
	      err,psline);

  width *= psunit;
  scalex = (lle-llw)/width;
  scaley = (lln-lls)/width;

  if (llw>lle || lls>lln)
    errorexit(3,"corner position error");

  prline(". southwest corner: %s",latlonstr(s,lls,llw));
  fprintf(stderr,"%s: southwest corner: %s\n",PROGNAME,s);
  prline(". northeast corner: %s",latlonstr(s,lln,lle));
  fprintf(stderr,"%s: northeast corner: %s\n",PROGNAME,s);
  prline(". size: %ld dots / resolution: %f m",
	 (long)(width+.5),max(scalex,scaley)*60.*NAUT_MILE);
  fprintf(stderr,"%s: size: %ld dots / resolution: %f m\n",
	  PROGNAME,(long)(width+.5),max(scalex,scaley)*60.*NAUT_MILE);
  prline(". skiprate: %d",skiprate);
  fprintf(stderr,"%s: skiprate: %d\n",PROGNAME,skiprate);

  /* data */
  section = "start data";
  while (1) {
    getl(s);
    if (*s!='%' && strstr(s," M"))
      break;
  }

  gmtline = gmtpoly = gmtpoint = 0;
  section = "data";
  state = 1;
  polyflag = 0;
  bufidx = 0;
  while (1) {
    if (*s!='%') {
      p = s;
      while (1) {
	while (*p==' ')
	  p++;
	if (!*p)
	  break;
	n = p;
	while ((*n>='0' && *n<='9') || *n=='-' || *n==' ')
	  n++;
	c = *n;
	if (*n)
	  *(n++) = '\0';
	if (c=='M') {
	  if (state==2)
	    bufwrite();
	  if (sscanf(p,"%ld %ld",&x,&y)!=2)
	    errorexit(4,"data M parse error (line %d)",psline);
	  bufadd(x,y);
	  state = 2;
	  p = n;
	}
	else if (c=='D') {
	  if (state!=2) {
	    state = 0;
	    break;
	  }
	  if (sscanf(p,"%ld %ld",&dx,&dy)!=2)
	    errorexit(4,"data D parse error (line %d)",psline);
	  x += dx;
	  y += dy;
	  bufadd(x,y);
	  p = n;
	}
	else if (c=='P') {
	  if (state!=2) {
	    state = 0;
	    break;
	  }
	  polyflag = 1;
	  p = n;
	}
	else if (c=='S') {
	  if (state!=2) {
	    state = 0;
	    break;
	  }
	  bufwrite();
	  state = 1;
	  p = n;
	}
	else if (c) {
	  state = 0;
	  break;
	}
      }
    }
    if (!state)
      break;
    getl(s);
  }

  fprintf(stderr,"%s: GMT:  %ld lines  %ld polys  (%ld points)\n",
	  PROGNAME,gmtline,gmtpoly,gmtpoint);
  fprintf(stderr,"%s: CHART:  %ld lines  %ld polys  (%ld points)\n",
	  PROGNAME,chaline,chapoly,chapoint);
}


int main (int argc, char *argv[])
{
  int a = 0;

  skiprate = 1;
  if (argc > 1 && *argv[1]=='-')
    a++;
  if ((a && sscanf(argv[1]+1,"%d",&skiprate)!=1) || argc>a+3) {
    fprintf(stderr,
	    "Usage: %s [-<skiprate>] [<gmtchart.ps> [<gmtchart.cha>]]\n"
	    "  default skiprate: 1 (i.e. don't skip data points)\n"
	    "GMT map of w/e/s/n area (west/east/south/north) with dataset r\n"
	    "  pscoast -JX1000d -Rw/e/s/n -Dr -W | %s > gmtchart.cha\n"
	    "Example: world map (crude dataset) / use every 10th data point\n"
	    "  pscoast -JX1000d -R-180/180/-90/90 -Dc -W | %s -10 > gmtworld.cha\n"
	    "Example: demo map (full dataset) / use all data points\n"
	    "  pscoast -JX1000d -R10:20/10:35/59:25/59:35 -Df -W | %s > gmtdemo.cha\n",
	    PROGNAME,PROGNAME,PROGNAME,PROGNAME);
    exit(-1);
  }
  if (argc >= a+2) {
    if (!(in=fopen(argv[a+1],"r")))
      errorexit(-1,"error opening input file %s\n",argv[a+1]);
  }
  else
    in = stdin;
  if (argc == a+3) {
    if (!(out=fopen(argv[a+2],"w")))
      errorexit(-1,"error opening output file %s\n",argv[a+2]);
  }
  else
    out = stdout;

  convert();

  if (in!=stdin)
    fclose(in);
  if (out!=stdout)
    fclose(out);

  return 0;
}
